import pytest

from composeml import LabelTimes
from composeml.tests.utils import read_csv, to_csv


@pytest.fixture
def labels(labels):
    return labels.threshold(100)


def test_sample_n_int(labels):
    given_answer = labels.sample(n=2, random_state=0)
    given_answer = given_answer.sort_index()
    given_answer = to_csv(given_answer, index=True)

    answer = [
        "label_id,customer_id,time,my_labeling_function",
        "2,2,2014-01-01 00:01:00,True",
        "3,2,2014-01-01 00:04:00,False",
    ]

    assert given_answer == answer


def test_sample_n_per_label(labels):
    n = {True: 1, False: 2}
    given_answer = labels.sample(n=n, random_state=0)
    given_answer = given_answer.sort_index()
    given_answer = to_csv(given_answer, index=True)

    answer = [
        "label_id,customer_id,time,my_labeling_function",
        "1,1,2014-01-01 00:48:00,False",
        "2,2,2014-01-01 00:01:00,True",
        "3,2,2014-01-01 00:04:00,False",
    ]

    assert given_answer == answer


def test_sample_frac_int(labels):
    given_answer = labels.sample(frac=0.25, random_state=0)
    given_answer = given_answer.sort_index()
    given_answer = to_csv(given_answer, index=True)

    answer = [
        "label_id,customer_id,time,my_labeling_function",
        "2,2,2014-01-01 00:01:00,True",
    ]

    assert given_answer == answer


def test_sample_frac_per_label(labels):
    frac = {True: 1.0, False: 0.5}
    given_answer = labels.sample(frac=frac, random_state=0)
    given_answer = given_answer.sort_index()
    given_answer = to_csv(given_answer, index=True)

    answer = [
        "label_id,customer_id,time,my_labeling_function",
        "0,1,2014-01-01 00:45:00,True",
        "2,2,2014-01-01 00:01:00,True",
        "3,2,2014-01-01 00:04:00,False",
    ]

    assert given_answer == answer


def test_sample_in_transforms(labels):
    n = {True: 2, False: 2}

    transform = {
        "transform": "sample",
        "n": n,
        "frac": None,
        "random_state": None,
        "replace": False,
        "per_instance": False,
    }

    sample = labels.sample(n=n)
    assert transform != labels.transforms[-1]
    assert transform == sample.transforms[-1]


def test_sample_with_replacement(labels):
    assert labels.shape[0] < 20
    n = {True: 10, False: 10}
    sample = labels.sample(n=n, replace=True)
    assert sample.shape[0] == 20


def test_single_target(total_spent):
    lt = total_spent.copy()
    lt.target_columns.append("target_2")
    match = "must first select an individual target"
    with pytest.raises(AssertionError, match=match):
        lt.sample(2)


def test_sample_n_per_instance():
    data = read_csv(
        [
            "target_dataframe_index,labels",
            "0,a",
            "0,b",
            "1,a",
            "1,b",
        ]
    )

    lt = LabelTimes(data=data, target_dataframe_index="target_dataframe_index")
    sample = lt.sample(n={"a": 1}, per_instance=True, random_state=0)
    actual = to_csv(sample, index=False)

    expected = [
        "target_dataframe_index,labels",
        "0,a",
        "1,a",
    ]

    assert expected == actual


def test_sample_frac_per_instance():
    data = read_csv(
        [
            "target_dataframe_index,labels",
            "0,a",
            "0,a",
            "0,a",
            "0,a",
            "1,a",
            "1,a",
        ]
    )

    lt = LabelTimes(data=data, target_dataframe_index="target_dataframe_index")
    sample = lt.sample(frac={"a": 0.5}, per_instance=True, random_state=0)
    actual = to_csv(sample, index=False)

    expected = [
        "target_dataframe_index,labels",
        "0,a",
        "0,a",
        "1,a",
    ]

    assert expected == actual



from collections import Counter

from pandas import isnull


class ExampleSearch:
    """A label search based on the number of examples.

    Args:
        expected_count (int): The expected number of examples to find.
    """

    def __init__(self, expected_count):
        self.expected_count = self._check_number(expected_count)
        self.reset_count()

    @staticmethod
    def _check_number(n):
        """Checks and formats the expected number of examples."""
        if n == -1 or n == "inf":
            return float("inf")
        else:
            info = "expected count must be numeric"
            assert isinstance(n, (int, float)), info
            return n

    @staticmethod
    def _is_finite_number(n):
        """Checks if a number if finite."""
        return n > 0 and abs(n) != float("inf")

    @property
    def is_complete(self):
        """Whether the search has found the expected number of examples."""
        return self.actual_count >= self.expected_count

    @property
    def is_finite(self):
        """Whether the expected number of examples is a finite number."""
        return self._is_finite_number(self.expected_count)

    def is_valid_labels(self, labels):
        """Whether the label values are not null."""
        return not any(map(isnull, labels.values()))

    def reset_count(self):
        """Reset the internal count of actual labels."""
        self.actual_count = 0

    def update_count(self, labels):
        """Update the internal count of actual labels."""
        self.actual_count += 1


class LabelSearch(ExampleSearch):
    """A label search based on the number of examples for each label.

    Args:
        expected_label_counts (dict): The expected number of examples to be find for each label.
            The dictionary should map a label to the number of examples to find for the label.
    """

    def __init__(self, expected_label_counts):
        items = expected_label_counts.items()
        self.expected_label_counts = Counter(
            {label: self._check_number(count) for label, count in items}
        )
        self.expected_count = sum(self.expected_label_counts.values())
        self.actual_label_counts = Counter()

    @property
    def is_complete(self):
        """Whether the search has found the expected number of examples for each label."""
        return len(self.expected_label_counts - self.actual_label_counts) == 0

    def is_complete_label(self, label):
        """Whether the search has found the expected number of examples for a label."""
        return (
            self.actual_label_counts.get(label, 0) >= self.expected_label_counts[label]
        )

    def is_valid_labels(self, labels):
        """Whether label values meet the search criteria.

        The search criteria is defined as label values that are not null, expected by the user, and have not reached the expected count.
        When these conditions are met by any label value, the labels are set to return to the user.
        This includes the other label values which share the same cutoff time.

        Args:
            labels (dict): The actual label values found during a search.

        Returns:
            value (bool): The value is True when valid, otherwise False.
        """
        label_values = labels.values()
        not_null = super().is_valid_labels(labels)
        is_expected = not_null and any(
            label in self.expected_label_counts for label in label_values
        )
        value = is_expected and any(
            not self.is_complete_label(label) for label in label_values
        )
        return value

    def reset_count(self):
        """Reset the internal count of actual labels."""
        self.actual_label_counts.clear()

    def update_count(self, labels):
        """Update the internal count of the actual labels.

        Args:
            labels (dict): The actual label values found during a search.
        """
        self.actual_label_counts.update(labels.values())



import featuretools as ft
import pytest

from composeml import LabelMaker


def total_spent(df):
    total = df.amount.sum()
    return total


@pytest.fixture
def labels():
    df = ft.demo.load_mock_customer(return_single_table=True, random_seed=0)
    df = df[["transaction_time", "customer_id", "amount"]]
    df.sort_values("transaction_time", inplace=True)

    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="transaction_time",
        labeling_function=total_spent,
        window_size="1h",
    )

    lt = lm.search(
        df,
        minimum_data="10min",
        num_examples_per_instance=2,
        gap="30min",
        drop_empty=True,
        verbose=False,
    )

    lt = lt.threshold(1250)
    return lt


def test_dfs(labels):
    target_column = labels.target_columns[0]
    es = ft.demo.load_mock_customer(return_entityset=True, random_seed=0)
    feature_matrix, _ = ft.dfs(
        entityset=es,
        target_dataframe_name="customers",
        cutoff_time=labels,
        cutoff_time_in_index=True,
    )
    assert target_column in feature_matrix

    columns = ["customer_id", "time", target_column]
    given_labels = feature_matrix.reset_index()[columns]
    given_labels = given_labels.sort_values(["customer_id", "time"])
    given_labels = given_labels.reset_index(drop=True)
    given_labels = given_labels.rename_axis("label_id")

    assert given_labels.equals(labels)



from pytest import raises

from composeml.data_slice.offset import DataSliceOffset


def test_numeric_typecast():
    assert int(DataSliceOffset("1 nanosecond")) == 1
    assert float(DataSliceOffset("1970-01-01")) == 0.0


def test_numeric_typecast_errors():
    match = "offset must be position or frequency based"
    with raises(TypeError, match=match):
        int(DataSliceOffset("1970-01-01"))

    match = "offset must be a timestamp"
    with raises(TypeError, match=match):
        float(DataSliceOffset("1 nanosecond"))


def test_invalid_value():
    match = "offset must be position or time based"
    with raises(AssertionError, match=match):
        DataSliceOffset(None)


def test_alias_phrase():
    phrase = "until start of next month"
    actual = DataSliceOffset(phrase).value
    expected = DataSliceOffset("MS").value
    assert actual == expected

    phrase = "until start of next year"
    actual = DataSliceOffset(phrase).value
    expected = DataSliceOffset("YS").value
    assert actual == expected
